## $Id: Help.pm 1137 2012-01-09 00:19:39Z zw $

package Dursamvad::Help;
use utf8;
use Dursamvad::HelpParser;
use Tk;
use Tk::ROText;

# The constructor takes the program title and the main window
sub new {
my ($this, $title, $mw) = @_;
my $class = ref($this) || $this;
my $self = bless {
           -Title => $title,
           -MainWindow => $mw,
           -History => [],
           -Location => 0,
           -Displayed => '',
           -Keep => 0,
           }, $class;
return $self;
}

# Previous view, history is stored backwards, the most recentis at Location=0
sub previous {
my $self = shift;
my $hist = $self->{-History};
my $loc = $self->{-Location};
return unless $loc < scalar @$hist;
$self->{-Location} = ++$loc;
$self->{-Keep} = 1;
$self->help(@{$hist->[$loc]});
}

# Next view
sub next {
my $self = shift;
my $hist = $self->{-History};
my $loc = $self->{-Location};
return unless $loc > 0;
$self->{-Location} = --$loc;
$self->{-Keep} = 1;
$self->help(@{$hist->[$loc]});
}

# Response to a mouse click in the text widget
sub a_href {
my ($self, $w) = @_;
my $e = $w->XEvent;
my ($x, $y) = ($e->x, $e->y);
my $c = $w->index("\@$x,$y");
my ($A, $Z) = $w->tagPrevrange('a', $c);
my ($a, $z) = $w->tagNextrange('href', $A, $Z);
if ($a) {
  $self->help($w->get($a, $z));
} else {
  $w->messageBox(-icon => 'error', -message => 'Missing link.');
}}

# Main help function, gets the link string
sub help {
my ($self, $href, $id) = @_;
if ($href =~ /#/) {
  $href = $`;  $id = $';
}
unless ($self->{-Keep}) {
  my $loc = $self->{-Location};
  while ($loc) {
    shift @{$self->{-History}};  --$loc;
  }
  $self->{-Location} = 0;
}
$self->{-Keep} = 0;
$href = $self->{-Displayed} unless $href;
# Create the window unless it exists
if (exists $self->{-HelpWindow}) {
  $self->{-HelpWindow}->withdraw;
} else {
  my %ButtonOptions = (-width => 10, -borderwidth => 3);
  my %GridOptions = (-sticky => 'nsew', -padx => 3, -pady => 3);
  my $hw = $self->{-HelpWindow} = $self->{-MainWindow}->Toplevel(
                -title => "Help for " . $self->{-Title});
  $hw->withdraw;
  $hw->geometry('-3+0');
  my $tx = $self->{-TextWidget} = $hw->Scrolled('ROText', -wrap => 'word', -padx => 9,
                -width => 50, -height => 40, -scrollbars => 'osoe')->grid(
                '-', '-', '-', %GridOptions);
  $self->{-Contents} = $hw->Button(-text => 'Contents',
          -command => sub {$self->help('Contents')}, %ButtonOptions)->grid(
  $self->{-Previous} = $hw->Button(-text => 'Previous', -state => 'disabled',
          -command => sub {$self->previous}, %ButtonOptions),
  $self->{-Next} = $hw->Button(-text => 'Next', -state => 'disabled',
          -command => sub {$self->next}, %ButtonOptions),
  $self->{-Close} = $hw->Button(-text => 'Close', -command => sub {$hw->withdraw}, %ButtonOptions), %GridOptions);
  $self->{-StatusText} = '';
  $self->{-Status} = $hw->Label(-textvariable => \$self->{-StatusText}, -padx => 6,
          -relief => sunken, -fg => 'brown', -anchor => 'w')->grid('-', '-', '-', %GridOptions);
  my ($columns, $rows) = $hw->gridSize;
  for (my $i = 0; $i < $columns; $i++) {
    $hw->gridColumnconfigure($i, -weight => 1);
  }
  $hw->gridRowconfigure(0, -weight => 1);
  # Tags
  $tx->tagConfigure('header', -spacing1 => 7, -justify => 'center', -background => 'cyan');
  $tx->tagConfigure('p', -spacing1 => 3, -spacing3 => 2, -lmargin1 => 15);
  $tx->tagConfigure('em', -foreground => 'blue');
  $tx->tagConfigure('pre', -lmargin1 => 0, -spacing1 => 0, -spacing3 => 0, -wrap => 'none');
  $tx->tagConfigure('a', -foreground => 'red');
  $tx->tagConfigure('href', -elide => 1);
  $tx->tagBind('a', '<Button-1>', sub {$self->a_href(shift)});
  $tx->tagBind('a', '<Any-Enter>', sub {shift->configure(-cursor => 'hand2')});
  $tx->tagBind('a', '<Any-Leave>', sub {shift->configure(-cursor => 'xterm')});
  # Allowed tags
  $self->{-AllowedTags} = [qw/header p em pre a/];
  # Button binding
  $hw->bind('<Control-w>', sub {$hw->withdraw});
  $hw->bind('<F1>', sub {$self->help('Help')});
  foreach (['-Contents', 'Show the help contents'],
           ['-Previous', 'Go to the previous view in the history'],
           ['-Next', 'Go to the next view in the history'],
           ['-Close', 'Close the help window']) {
    my ($bname, $btext) = @$_;
    my $bref = $self->{$bname};
    $bref->bind('<Any-Enter>', sub {$self->{-StatusText} = $btext});
    $bref->bind('<Any-Leave>', sub {$self->{-StatusText} = ''});
  }
}

my $txt = $self->{-TextWidget};
if ($href ne $self->{-Displayed}) {
  my $Parser = new Dursamvad::HelpParser ($txt, $self->{-AllowedTags});
  $txt->delete('1.0', 'end');
  $self->show($Parser, $href);
  $self->{-Displayed} = $href;
}
if ($id) {
  $txt->see($id) if $txt->markExists($id);
} else {
  $txt->see('1.0');
}
my $hist = $self->{-History};
if ($self->{-Location} == 0) {
  if (scalar @$hist) {
    if ($hist->[0]->[0] ne $href || $hist->[0]->[1] ne $id) {
      unshift @$hist, [$href, $id];
    }
  } else {
    unshift @$hist, [$href, $id];
  }
  while (scalar @$hist > 99) {
    my $nic = shift @$hist;
  }
}
$self->{-Next}->configure(-state => $self->{-Location} > 0 ? 'normal' : 'disabled');
$self->{-Previous}->configure(-state => $self->{-Location} < scalar(@$hist)-1 ? 'normal' : 'disabled');
$self->{-HelpWindow}->deiconify;  $self->{-HelpWindow}->raise;
}

# Show the help page, expects the parser and the href
sub show {
my ($self, $Parser, $href) = @_;
my $string = "<p><em>No help for $href</em></p>";

$href eq 'Contents' and $string = <<ZZZ;
<header>Contents</header>
<p><a href='Help'>Using help</a></p>
<p><a href='About'>About the program</a></p>
<p><a href='Main'>Main window</a></p>
<p><a href='Terminal_setup'>Setting the port for the terminal mode</a></p>
<p><a href='Terminal'>Terminal mode</a></p>
<p><a href='Monitor_setup'>Setting the port for the monitor mode</a></p>
<p><a href='Monitor'>Monitor mode</a></p>
<p><a href='otherdoc'>Further documentation</a></p>
<p><a href='Problems'>Problems</a></p>
<p><a href='lic'>Author and license</a></p>
<!-- History is for debugging only! -->
<!-- p><a href='History'>History</a> for debugging only</p -->
ZZZ

$href eq 'lic' and $string = <<ZZZ;
<header>Author and license</header>
<p>The author of the program is Zdeněk Wagner – Ice Bear Soft, the web page of the program is:
<em>http://icebearsoft.euweb.cz/dursamvad/</em>.
The program can be used, modified and redistributed under the GNU General Public License version 2
or newer.</p>
ZZZ

$href eq 'About' and $string = <<ZZZ;
<header>About the program</header>

<p>Some measurement devices still communicate using the RS232 interface. A few years ago RS232 port
appeared on all computers. Recently USB ports are preferred.</p>

<p>Measurement devices do not always contain communication software. It is useful to have
possibility to test communication manually using a terminal program. Programs as <em>minicom</em>
and <em>picocom</em> are helpful but do not display the modem lines. Program <em>cutecom</em> does
not allow to use USB (at least the version distributed with Linux installed on my computer). I
wrote this simple program in order to solve these difficulties and in addition to learn how to
develop a terminal with GUI.</p>

<p>The program serves as a simple testing tool for RS232 communication. It can display not only the
received bytes but can also show the status of the modem lines as well as monitor communication of
two devices. It is possible to use USB to RS232 adapters.</p>

<p>The program was written and tested on Linux and may hopefully work on other Unix systems. It will
not work on other operating systems.</p>

<p>The program requires perl-Tk and Device::SerialPort. Both modules can be installed from CPAN. As
the first option you should use the packages included in your Linux distribution, try CPAN only if
the packages are not available. Specifically, perl-Tk on CPAN often requires newer perl than that
distributed with server-type Linux distros as RHEL or CentOS. Perl-Tk from CPAN may not work
properly in these distros while the rpm module, that can be installed and/or updated by yum, works
fine. Installation details are explained in the supplied <em>README</em> file.</p>
ZZZ

$href eq 'Main' and $string = <<ZZZ;
<header>The main window</header>

<p>The main window is the entry point to the program. It contains the <em>Terminal mode</em> and
<em>Monitor mode</em> buttons for selecting a mode, the <em>Exit</em> button for exiting the
program, and the <em>Help</em> button for displaying the help (you have found it, haven't you?).</p>

<p>The <a href='Terminal'>terminal mode</a> is intended for human driven communication of a computer
with a device (it may be another computer). The aim of the program is to allow testing and
debugging simple communication or to serve as a very simple terminal. It should not be considered a
full featured fast terminal. The user may type strings and see response.</p>

<p>The <a href='Monitor'>monitor mode</a> can be used to monitor communication of two devices.</p>

<header id='shortcuts'>Key bindings</header>

<p>Some actions can also be invoked by keyboard shortcuts. The following keys are defined:</p>

<p/>
<pre>
	Ctrl-q	exit the program
	Ctrl-w	close the window
	Ctrl-a	select all text in the text window
		(in the <a href='Terminal'>terminal</a> or <a href='Monitor'>monitor</a> mode)
	F1	show help
</pre>

<p>In the main window <em>Ctrl-W</em> exits the program. In the help window <em>F1</em> displays
information on using help. <em>Ctrl-Q</em> is not active in the help window.</p>

<p>There are <a href='Terminal#shortcuts'>additional shortcuts</a> defined for the terminal mode
only.</p>
ZZZ

$href eq 'Terminal_setup' and $string = <<ZZZ;
<header>Setting the port for the terminal mode</header>

<p>This dialog allows you to select the port and set its parameters. The selected values will then
be used in the <a href='Terminal'>Terminal mode</a>. When the dialog is opened, the program reads
the list of currently available ports. Remember that RS232 to USB adapters are not supported at
system sturtup but are added dynamically by hotplug which takes some time. If you do not see such a
device, you have to cancel the dialog, wait a short time and try again. You can also issue a
command <em>dmesg</em> from a terminal window in order to see whether hotplug has already found and
configured the device.</p>

<p>Whenever a new port is selected, the program reads the list of supported options and select the
current value. If the boxes below show empty values, it may have two reasons. Either the driver is
present but the hardware is missing or you do not have access rights for the device. See section <a
href='Problems'>Problems</a> for explanation. The program does not allow to select to high or too
low baudrates because there is a high risk of data loss. If you really need such a baudrate, you
have to use another program.</p>

<p><em>Important:</em> The program can only work if you set the same options as are set in the
device with which you wish to communicate.</p>
ZZZ

$href eq 'Monitor_setup' and $string = <<ZZZ;
<header>Setting the port for the monitor mode</header>

<p>This dialog is used to select ports and their options for the <a href='Monitor'>Monitor
mode</a>. Monitoring requires two distinct interfaces, therefore you have to choose two ports.
Selection is done the same way as for the <a href='Terminal_setup'>Terminal mode settings</a>.
Additionally you have to choose whether an input modem line level will be sent to a signal of the
other interface. If you send two or more lines to a single line of the other interface, logical OR
will be used. It is allowed to use asymetric settings. Moreover, the interfaces may use different
baudrates and different options. The only requiremt is that the ports must be different. I
handhake is set to <em>rts</em>, the port will not accept modem lines setting from the other port
no matter what you select in the dialog.</p>
ZZZ

$href eq 'Terminal' and $string = <<ZZZ;
<header>Terminal mode</header>

<p>The terminal mode is used to communicate with a serial device. At the top of the window the port
and settings are displayed as they were chosen in the <a href='Terminal_setup'>settings dialog</a>.
The second line shows the status of the modem lines.</p>

<p>The upper text area displays the incoming bytes. You can clear the window or log the input as it
arrives. If the input is being logged, the file name will appear in the area close to the button.
Remember that it is not possible to append data to an existing file. If you agree to overwrite an
existing file, all its contents will be lost.</p>

<p>Logging writes the data as they arrive. It is not possible to save the data already displayed.
However, you can copy&amp;paste them to another application.</p>

<p>The lower text window shows the history of your input. You can clear it or log your input.</p>

<p>Buttons <em>DTR</em> and <em>RTS</em> allow you to change the modem lines. Depending on your
selection they work either in the on/off mode or in a pulse mode. If you switch to the pulse mode,
the event loop will be locked during the pulse. This may have bad consequences if the baudrate is
high and the pulse time is long. The result may be data loss. In addition, if you are just testing
the interface by a loopback connector, you will not see the pulse at all.</p>

<p>The entry field is used to write the text strings that should be sent to the serial port. You
can select what will be appended to the string. Teh default is <em>LF</em>, other possibilities are
<em>none</em>, <em>CR</em>, and <em>CRLF</em>. The string is send when you press the <em>Enter</em>
key or the <em>Send</em> button. The strings are stored in the history. You can therefore use the
drop-down to select a string from a history, edit it and send it.</p>

<p id='sendfile'>The <em>Send file</em> button allows you to send a file. The file contents will not be shown in
the user input window. The file should be short and contain only ASCII characters. Timing is set not to
lose data but there is no guarantee. Binary data may lock the port mainly when <em>xon/xoff</em>
handshake is used.</p>

<p>The <em>Close</em> button closes the window and returns to the main window, the <em>Exit</em>
button exits the program.</p>

<header id='shortcuts'>Additional shortcuts</header>

<p>The terminal mode offers three additional hot keys. <em>Ctrl-t</em> sends the current local
time, <em>Ctrl-g</em> sends the current UTC time. Both are sent immediatelly in the ISO 8601 format
and are not stored in the history. They are displayed in the user input text window and will be
logged if logging is enabled. <em>Ctrl-u</em> fills the URL of this program to the entry field. If
you really wish to send it, you have to press <em>Enter</em> or <em>Send</em>. In such a case it
will appear in the history.</p>

<p><em>Important warning:</em> Do not log both the port input and user input to the same file in
the same time. The program will not detect it and will probably crash sooner or later.</p>
ZZZ

$href eq 'Monitor' and $string = <<ZZZ;
<header>Monitor mode</header>

<p>This mode is intended for monitoring communication between two serial devices. It is a passive
monitor composed of two port input windows with functions similar as in the <a
href='Terminal'>terminal mode</a>. The only active operation available is logging any port
input.</p>

<p><em>Important warning:</em> Do not log input from both ports to the same file in
the same time. The program will not detect it and will probably crash sooner or later.</p>
ZZZ

$href eq 'otherdoc' and $string = <<ZZZ;
<header>Additional documentation</header>

<p>Additional documentation is available in the POD form in the main file, <em>dursamvad.pl</em>. It
covers details not explained in this help. The documentation can be viewed by <em>perldoc</em>.
Use:</p>

<pre>
perldoc perldoc
</pre>

<p>if you do not know how to use it. Notice that the documentation can be displayed even if the
program has not yet been installed.</p>

<p>Installation notes are also available from the <em>README</em> file.</p>
ZZZ

$href eq 'Problems' and $string = <<ZZZ;
<header>Problems</header>

<p><em>Problem:</em> The RS232 to USB adapter device <em>/dev/tty/USBn</em> is not offered in the
list of serial interfaces.</p>

<p><em>Solution:</em> The device is created by a hotplug. You have opened the settings dialog too
early. Cancel the dialog, wait a short time and try again. It is described at the <a
href='Terminal_setup'>terminal port setting</a> help page.</p>

<p><em>Problem:</em> Options for all serial devices are blank.</p>

<p><em>Solution:</em> You do not have proper access rights. The devices are owned by <em>root</em>
and a special group the name of which is system/distribution dependent. The group name can be
displayed by <em>ls -l /dev/tty*</em>. In order to have access rights you must be member of that
group. You will then be granted the access rights on the next login.</p>

<p><em>Problem:</em> Options for some serial devices are blank.</p>

<p><em>Solution:</em> Your system offers a device for a nonexistent hardware which is quite common.
In principle four serial devices can be present but usually just one or two exist in a PC. Some
computers do not have RS232 devices at all. The only remedy is to use existing hardware.</p>

<p><em>Problem:</em> A limited number of baudrates are offered, I need higher or lower
baudrate.</p>

<p><em>Solution:</em> The range of baudrates is limited in order to make the communication
reliable. With a too high baudrate the program does not manage to read the characters from the
input buffer suficiently quickly, with a too low baudrate the program may be so fast that the
bytes are not sent with a sufficient speed. In both cases characters are lost. The offered range
results from reliability testing. If you need a baudrate outside the offered values, you have to
use another program.</p>

<p><em>Problem:</em> The strings with accented characters seem to be sent to a port, are properly
remembered in the history and properly shown in the user input text window but if such characters
are input from the port, garbage is displayed.</p>

<p><em>Solution:</em> The program assumes strings in UTF-8 which is a multibyte encoding but the
serial ports works with bytes. In order to guarantee reliability the program tries to process input
bytes as fast as possible. It may therefore happen that the bytes string is not valid UTF-8 string
although UTF-8 text is expected. Attempts to handle the input bytes as if they were UTF-8 strings
lead to even worse reliability, therefore the port input is simply consider as bytes. There is no
possibility to treat them as anything else than ASCII characters.</p>

<p><em>Problem:</em> When sending a binary file the program locks and it is only possible to close
the window.</p>

<p><em>Solution:</em> The program is not prepared for sending binary files. The file may even
contain control characters that will pause the serial port. If you wish to work with binary files,
you have to use another program. More information can be found in the description of the <a
href='Terminal#sendfile'>Send file</a> button.</p>

<p><em>Problem:</em> It is not possible to send the RI signal to the other port in the monitor
mode.</p>

<p><em>Solution:</em> The RI line is the input signal in the computer. The computer can accept it
but cannot send it. You have to prepare a special cable, or even a switchdesk connecting some lines
at a hardware level. The connection depends on the devices you wish to monitor, no general rules
can be given.</p>

<p><em>Problem:</em> The devices do not communicate if handshake is set to <em>rts</em>.</p>

<p><em>Solution:</em> It is not clear how hanshaking is handled in the driver. The program assumes
that the kernel drivers or even hardware do all what is needed. If you find an authoritative
document saying what has to be done at the application level, the author will fix it in the next
release. See the page <a href='lic'>Author and license</a> for the URL of the program containing
the link to a contact email.</p>
ZZZ

$href eq 'Help' and $string = <<ZZZ;
<header>Using help</header>

<p>The help window displays information related to a particular dialog. You can keep it open while
you work with the program or you can close it in order to get more space on the screen.</p>

<p>The information may provide hyperlinks to other pages. The hyperlinks are displayed with red and
the mouse cursor changes when over them. If you follow a hyperlink, you may then
use the <em>Previous</em> and <em>Next</em> buttons to move along the browsing history. The
<em>Contents</em> button will navigate you to the contents page.</p>

<p>No matter whether the help window is shwn or hidden, the help for the current dialog may be
displayed by pressing the <em>Help</em> button or the <em>F1</em> key.</p>

<p><em>F1</em> is just one of the available shortcuts. You can see the <a href='Main#keys'>list of
shortcuts</a>.</p>
ZZZ

$href eq 'History' && do {
  $string = '<p id="end"><em>End of history</em></p>';
  $Parser->parse('<header>History</header><p><a href="#end">Go to end</a></p>');
  foreach (@{$self->{-History}}) {
    my $url = $_->[0];
    $url .= '#' . $_->[1] if $_->[1];
    $Parser->parse("<p>$url</p>");
  }
};

$Parser->parse($string);
}

1;

__END__

=head1 NAME

Dursamvad::Help

=head1 SYNOPSIS

Help displaying class designed for the DURSAMVAD program. It requires C<Dursamvad::Parser> and
C<Dursamvad::HelpParser> packages.

=head1 DESCRIPTION

This class implements the help display including support for clickable hyperlinks. The help is
written in an XML-like form but without the document tag. All tags must be configured in the text
widget. The @id attribute is used to set a mark that may be a target of a hyperlink. The hyperlink
is a string of a form C<name#id> where C<name> is the name of the help page. Either part is
optional, the hyperlink thus may also be in the form of C<name> or C<#id>.

=head2 new

The constructor creates an instance that should be the only instance living throughout the whole
program lifetime. Its reference should therefore be stored in a global variable. The constructor
requires the program title and the reference to the main window.

=head2 previous

This is the internal function that shows the previous view in the history.

=head2 next

This is the internal function for showing the next view in the history (after I<previous> has been
used).

=head2 a_href

This internal function responds to mouse clicks inside the help display.

=head2 help

This function does all the job. The Toplevel window is created only once. Several tags are
configured for use with the text widget. If the window is closed, it is only withdrawn but not
destroyed. When help is invoked again, the same window is reused and its last geometry is
remembered. The function calls C<show> to insert the contents to the text window.

=head2 show

This function gets the parser reference and the name. It should yse the parser's C<parse> function
to add the contents.

=head1 USAGE

The program should create a global instance and call help from buttons via anonymous subroutines.
Help should send a single parameter in the form of C<name> or C<name#id>. Key bindings should be
defined in a similar way. This is a very simple example:

	our $MainWindow = new MainWindow;
	our $Help = new Dursamvad::Help ('Title', $MainWindow);
	$MainWindow->Button(-text => 'Help', -command => sub {$Help->help('main_window')})->pack();
	$MainWindow->bind('<F1>', sub {$Help->help('main_window')});
	...
	MainLoop;

It is, of course, also possible to bind specific help events to buttons and other controlls. They
may for instance jump to a mark in the window as in the example below:

	our $button = $MainWindow->Button(-text => 'Something', -command => \&doSomethink)->pack();
	$button->bind('<F1>', sub {$Help->help('main_window#something')});

=head1 LICENSE

The package was developed as a part of the DURSAMVAD program, see C<dursamvad.pl> for further info.

=cut


